
from unittest import TestCase

from tololib.command_data_type import Boolean, Integer, OptionalInteger, IntegerEnum
from tololib.const import LampMode


class BooleanTest(TestCase):
    def test_to_typed(self) -> None:
        cdt = Boolean()
        self.assertEqual(cdt.to_typed(True), True)
        self.assertEqual(cdt.to_typed('true'), True)
        self.assertEqual(cdt.to_typed('1'), True)
        self.assertEqual(cdt.to_typed('on'), True)
        self.assertEqual(cdt.to_typed('yes'), True)

        self.assertEqual(cdt.to_typed(False), False)
        self.assertEqual(cdt.to_typed('false'), False)
        self.assertEqual(cdt.to_typed('0'), False)
        self.assertEqual(cdt.to_typed('off'), False)
        self.assertEqual(cdt.to_typed('no'), False)

        self.assertRaises(ValueError, cdt.to_typed, 'foobar')

    def test_to_byte(self) -> None:
        cdt = Boolean()
        self.assertEqual(cdt.to_byte(True), 1)
        self.assertEqual(cdt.to_byte(False), 0)


class IntegerTest(TestCase):
    def test_to_typed(self) -> None:
        cdt = Integer(10, 90)
        self.assertEqual(cdt.to_typed(42), 42)
        self.assertEqual(cdt.to_typed('42'), 42)

        self.assertRaises(ValueError, cdt.to_typed, 9)

        self.assertRaises(ValueError, cdt.to_typed, None)
        self.assertRaises(ValueError, cdt.to_typed, 'foobar')


class OptionalIntegerTest(TestCase):
    def test_to_typed(self) -> None:
        cdt = OptionalInteger(1, 60, 61)
        self.assertEqual(cdt.to_typed(61), None)
        self.assertEqual(cdt.to_typed(42), 42)
        self.assertEqual(cdt.to_typed('42'), 42)

        self.assertRaises(ValueError, cdt.to_typed, 0)

    def test_to_byte(self) -> None:
        cdt = OptionalInteger(1, 60, 61)
        self.assertEqual(cdt.to_byte(None), 61)


class IntegerEnumTest(TestCase):
    def test_to_typed(self) -> None:
        cdt = IntegerEnum(LampMode)
        self.assertEqual(cdt.to_typed(LampMode.MANUAL), LampMode.MANUAL)
        self.assertEqual(cdt.to_typed(LampMode.AUTOMATIC), LampMode.AUTOMATIC)
        self.assertEqual(cdt.to_typed(0), LampMode.MANUAL)
        self.assertEqual(cdt.to_typed(1), LampMode.AUTOMATIC)
        self.assertEqual(cdt.to_typed('Manual'), LampMode.MANUAL)
        self.assertEqual(cdt.to_typed('automatic'), LampMode.AUTOMATIC)

        self.assertRaises(ValueError, cdt.to_typed, LampMode)

    def test_to_byte(self) -> None:
        cdt = IntegerEnum(LampMode)
        self.assertEqual(cdt.to_byte(LampMode.MANUAL), 0)
        self.assertEqual(cdt.to_byte(LampMode.AUTOMATIC), 1)




from unittest import TestCase

from tololib.const import Calefaction, Model
from tololib.message_info import StatusInfo, SettingsInfo


class StatusInfoTest(TestCase):
    def test_init(self) -> None:
        self.assertRaises(ValueError, StatusInfo, b'foobar')

    def test_getters_setters(self) -> None:
        info = StatusInfo()

        info.power_on = True
        self.assertEqual(info.power_on, True)
        self.assertEqual(info[0], 1)

        info.power_on = False
        self.assertEqual(info.power_on, False)
        self.assertEqual(info[0], 0)

        info.current_temperature = 45
        self.assertEqual(info.current_temperature, 45)
        self.assertEqual(info[1], 45)

        info.power_timer = None
        self.assertEqual(info.power_timer, None)
        self.assertEqual(info[2], 255)

        info.power_timer = 30
        self.assertEqual(info.power_timer, 30)
        self.assertEqual(info[2], 30)

        info.aroma_therapy_on = True
        self.assertTrue(info.aroma_therapy_on)
        self.assertEqual(info[4], 1)

        info.aroma_therapy_on = False
        self.assertFalse(info.aroma_therapy_on)
        self.assertEqual(info[4], 0)

        info.sweep_on = True
        self.assertTrue(info.sweep_on)
        self.assertEqual(info[5], 1)

        info.sweep_on = False
        self.assertFalse(info.sweep_on)
        self.assertEqual(info[5], 0)

        info.lamp_on = True
        self.assertTrue(info.lamp_on)
        self.assertEqual(info[7], 1)

        info.lamp_on = False
        self.assertFalse(info.lamp_on)
        self.assertEqual(info[7], 0)

        info.water_level = 2
        self.assertEqual(info.water_level, 2)
        self.assertEqual(info[8], 2)

        info.fan_on = True
        self.assertTrue(info.fan_on)
        self.assertEqual(info[9], 1)

        info.fan_on = False
        self.assertFalse(info.fan_on)
        self.assertEqual(info[9], 0)

        info.fan_timer = None
        self.assertIsNone(info.fan_timer)
        self.assertEqual(info[10], 61)

        info.fan_timer = 30
        self.assertEqual(info.fan_timer, 30)
        self.assertEqual(info[10], 30)

        info.current_humidity = 95
        self.assertEqual(info.current_humidity, 95)
        self.assertEqual(info[11], 95)

        info.tank_temperature = 23
        self.assertEqual(info.tank_temperature, 23)
        self.assertEqual(info[12], 23)

        info.model = Model.DOMESTIC
        self.assertEqual(info.model, Model.DOMESTIC)
        self.assertEqual(info[14], 0)

        info.model = Model.COMMERCIAL
        self.assertEqual(info.model, Model.COMMERCIAL)
        self.assertEqual(info[14], 1)

        info.salt_bath_on = True
        self.assertTrue(info.salt_bath_on)
        self.assertEqual(info[15], 1)

        info.salt_bath_on = False
        self.assertFalse(info.salt_bath_on)
        self.assertEqual(info[15], 0)

        info.salt_bath_timer = None
        self.assertIsNone(info.salt_bath_timer)
        self.assertEqual(info[16], 255)

        info.salt_bath_timer = 10
        self.assertEqual(info.salt_bath_timer, 10)
        self.assertEqual(info[16], 10)

    def test_flow_in_out_calefaction(self) -> None:
        info = StatusInfo()
        self.assertFalse(info.flow_in)
        self.assertFalse(info.flow_out)
        self.assertEqual(info.calefaction, 0)

        info.flow_in = True
        self.assertTrue(info.flow_in)
        self.assertFalse(info.flow_out)
        self.assertEqual(info.calefaction, 0)

        info.calefaction = Calefaction.KEEP
        self.assertTrue(info.flow_in)
        self.assertFalse(info.flow_out)
        self.assertEqual(info.calefaction, Calefaction.KEEP)

        info.flow_out = True
        self.assertTrue(info.flow_in)
        self.assertTrue(info.flow_out)
        self.assertEqual(info.calefaction, Calefaction.KEEP)

        info.calefaction = Calefaction.HEAT
        self.assertTrue(info.flow_in)
        self.assertTrue(info.flow_out)
        self.assertEqual(info.calefaction, Calefaction.HEAT)

        info.flow_in = False
        self.assertFalse(info.flow_in)
        self.assertTrue(info.flow_out)
        self.assertEqual(info.calefaction, Calefaction.HEAT)


class SettingsInfoTest(TestCase):
    def test_init(self) -> None:
        self.assertRaises(ValueError, SettingsInfo, b'foobar')

    def test_getters_setters(self) -> None:
        info = SettingsInfo()

        info.target_temperature = 42
        self.assertEqual(info.target_temperature, 42)
        self.assertEqual(info[0], 42)




import logging
import socket
from select import select
from threading import Thread
from typing import Any, Optional, Tuple, Union

from .command import Command
from .command_property_map import (
    SET_STATUS_COMMAND_PROPERTY_MAP, GET_STATUS_COMMAND_PROPERTY_MAP,
    SET_SETTINGS_COMMAND_PROPERTY_MAP, GET_SETTINGS_COMMAND_PROPERTY_MAP
)
from .message import Message
from .message_info import MessageInfo, StatusInfo, SettingsInfo


logger = logging.getLogger(__name__)


class ToloServer(object):
    def __init__(self, address: Optional[str] = None, port: int = 51500) -> None:
        self._address = address
        self._port = port

        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.bind((address, port))
        self._thread: Optional[Thread] = None
        self._keep_running = False

        self._status = StatusInfo()
        self._settings = SettingsInfo()

    def start(self) -> None:
        """Initialize and start the server loop thread.

        Initializes and starts a thread in the background, which handles incoming messages and updates local variables
        accordingly.
        """
        if self._thread is not None:
            raise RuntimeError('server thread already initialized!')

        self._thread = Thread(target=self._thread_loop)
        self._keep_running = True
        self._thread.start()

    def join(self, timeout: Optional[float] = None) -> None:
        """Join the background thread and wait until it is finished.

        Args:
            timeout (float): timeout in seconds
        """
        if self._thread is None:
            raise RuntimeError('server does not have a running thread - did you call start() before?')

        self._thread.join(timeout)

    def run(self) -> None:
        """Shortcut method calling start() and join()"""
        self.start()
        self.join()

    def receive_message(self) -> Tuple[Message, str]:
        """Receive, decode and return a single control message.

        Blocks until a message has received.

        Returns:
            A decoded Message object which contains the received control message.
        """
        raw_bytes, sender = self._socket.recvfrom(4096)
        return Message.from_bytes(raw_bytes), sender

    def send_message(self, message: Message, recipient: Union[Tuple[str, int], str]) -> int:
        raw_bytes = bytes(message)
        return self._socket.sendto(raw_bytes, recipient)

    @staticmethod
    def _set_and_get_info(message: Message, property_name: Optional[str], info: MessageInfo) -> Message:
        if property_name is None:
            raise ValueError('None value not allowed here')
        setattr(info, property_name, message.data)
        return Message(
            message.command,
            message.data,
            b'\x00'
        )

    @staticmethod
    def _get_info(message: Message, property_name: Optional[str], info: MessageInfo) -> Message:
        if property_name is None:
            raise ValueError('None value not allowed here')
        return Message(
            message.command,
            getattr(info, property_name),
            b'\x00'
        )

    def _thread_loop(self) -> None:
        logger.debug('server thread loop started')
        while self._keep_running:
            incoming_data = select([self._socket], [], [], 0.2)
            if incoming_data[0]:
                message, sender = self.receive_message()
                logger.debug('received message %s' % str(message))

                # status request

                if message.command == Command.STATUS:
                    self.send_message(Message(Command.STATUS, 0, self._status), sender)

                # status change commands

                elif message.command in SET_STATUS_COMMAND_PROPERTY_MAP.keys():
                    self.send_message(self._set_and_get_info(
                        message,
                        SET_STATUS_COMMAND_PROPERTY_MAP.get(message.command),
                        self._status
                    ), sender)

                # status getter commands

                elif message.command in GET_STATUS_COMMAND_PROPERTY_MAP.keys():
                    self.send_message(self._get_info(
                        message,
                        GET_STATUS_COMMAND_PROPERTY_MAP.get(message.command),
                        self._status
                    ), sender)

                # settings request

                elif message.command == Command.SETTINGS:
                    self.send_message(Message(Command.SETTINGS, 0, self._settings), sender)

                # settings change commands

                elif message.command in SET_SETTINGS_COMMAND_PROPERTY_MAP.keys():
                    self.send_message(self._set_and_get_info(
                        message,
                        SET_SETTINGS_COMMAND_PROPERTY_MAP.get(message.command),
                        self._settings
                    ), sender)

                # settings getter commands

                elif message.command in GET_SETTINGS_COMMAND_PROPERTY_MAP.keys():
                    self.send_message(self._get_info(
                        message,
                        GET_SETTINGS_COMMAND_PROPERTY_MAP.get(message.command),
                        self._settings
                    ), sender)

                else:
                    logger.warning('unhandled message: %s' % str(message))

    def stop(self) -> None:
        self._keep_running = False

    def close(self) -> None:
        self._socket.close()

    def get_socket_info(self) -> Any:
        return self._socket.getsockname()

    @property
    def status(self) -> StatusInfo:
        return self._status

    @property
    def settings(self) -> SettingsInfo:
        return self._settings

    def __del__(self) -> None:
        self.close()




from typing import Union

from .command import Command, CommandTypeHint
from .const import MESSAGE_PREFIX, MESSAGE_SUFFIX


class Message(object):
    """Class representing a message sent over network.

    By default, the TOLO App Box listens on UDP port 51500 for receiving control messages.
    Each messaged received by TOLO App Box is answered with a message with the same structure, but different values.
    This class represents such request/response messages and provides functionality for parsing, validating and
    creating messages.

    Each message has the following structure:
      * 2 bytes fixed `0xAAAA`
      * 1 byte command (see Command enumeration)
      * 1 byte data (according to the command)
      * x bytes response message (depends on the command and the reply from TOLO App Box)
      * 2 bytes fixed `0x5555`
      * 1 byte checksum, byte-wise XOR of all previously mentioned data
    """
    PREFIX = MESSAGE_PREFIX
    SUFFIX = MESSAGE_SUFFIX

    def __init__(self, command: Command, data: CommandTypeHint, info: Union[bytes, bytearray] = b''):
        command.command_data_type.validate(data, raise_error=True)

        self._command = command
        self._data = data
        self._info = bytes(info)

    @property
    def command(self) -> Command:
        return self._command

    @property
    def data(self) -> CommandTypeHint:
        return self._data

    @property
    def info(self) -> bytes:
        return self._info

    def __str__(self) -> str:
        return '<%s: cmd: %s, data: %d, info: %s>' % (
            self.__class__.__name__,
            self._command.name,
            self._command.command_data_type.to_byte_any(self._data),
            self._info.hex()
        )

    def __bytes__(self) -> bytes:
        data = self.PREFIX + bytes([
            self._command.code,
            self._command.command_data_type.to_byte(self._data)
        ]) + self._info + self.SUFFIX
        return data + bytes([self.generate_crc(data)])

    @staticmethod
    def from_bytes(b: bytes) -> 'Message':
        if not Message.validate_meta(b):
            raise ValueError('invalid meta information')

        command = Command.from_code(b[2])
        data = command.command_data_type.to_typed(b[3])

        return Message(
            command=command,
            data=data,
            info=b[4:-3]
        )

    @staticmethod
    def generate_crc(data: bytes) -> int:
        crc = 0x00
        for b in data:
            crc = crc ^ b
        return crc

    @staticmethod
    def validate_crc(raw_bytes: bytes) -> bool:
        return Message.generate_crc(raw_bytes[:-1]) == raw_bytes[-1]

    @classmethod
    def validate_meta(cls, raw_bytes: bytes) -> bool:
        """Validates meta data of message bytes.

        The validation will check for prefix, suffix and CRC.
        It will NOT check for valid code or payload.

        Args:
            raw_bytes (bytes): binary data to be checked

        Returns:
            True if the check was successful and the meta data is as expected, False otherwise.
        """
        if not raw_bytes.startswith(cls.PREFIX):
            return False
        if not raw_bytes[:-1].endswith(cls.SUFFIX):
            return False
        return cls.validate_crc(raw_bytes)



"""
TOLOlib is a Python Library for Controlling TOLO Sauna/Steam Bath Devices.

Note: This is a community project licensed under MIT license.
There is no professional affiliation between this project or the author of this project with companies behind TOLO.

The source code can be found at https://gitlab.com/MatthiasLohr/tololib.
The library package is published at PyPI here: https://pypi.org/project/tolosteambath/
"""

from .client import ToloClient
from .command import Command
from .message import Message
from .server import ToloServer


__all__ = [
    'Command',
    'Message',
    'ToloClient',
    'ToloServer'
]




from enum import IntEnum
from typing import Any, Generic, Optional, Type, TypeVar


E = TypeVar('E', bound=IntEnum)
T = TypeVar('T')


class CommandDataType(Generic[T]):
    def to_typed(self, value: Any) -> T:
        raise NotImplementedError()

    def validate(self, value: Any, raise_error: bool = True) -> bool:
        try:
            self.to_typed(value)
            return True
        except ValueError as e:
            if raise_error:
                raise e
            else:
                return False

    def to_byte(self, value: T) -> int:
        raise NotImplementedError()

    def to_byte_any(self, value: Any) -> int:
        return self.to_byte(self.to_typed(value))


class Boolean(CommandDataType[bool]):
    def to_typed(self, value: Any) -> bool:
        if isinstance(value, bool):
            return value
        elif isinstance(value, int):
            return bool(value)
        elif isinstance(value, str):
            if value.lower() in ('false', '0', 'off', 'no'):
                return False
            elif value.lower() in ('true', '1', 'on', 'yes'):
                return True
            else:
                raise ValueError('cannot interpret %s as boolean value' % value)
        else:
            raise ValueError('unsupported type for normalization')

    def to_byte(self, value: bool) -> int:
        return 1 if value else 0


class Integer(CommandDataType[int]):
    def __init__(self, _min: int, _max: int) -> None:
        self._min = _min
        self._max = _max

    def to_typed(self, value: Any) -> int:
        if isinstance(value, int):
            return self._ensure_min_max(value)
        elif isinstance(value, str):
            return self._ensure_min_max(int(value))
        else:
            raise ValueError('cannot parse %s to int' % type(value))

    def to_byte(self, value: int) -> int:
        while value < 0:
            value += 256
        while value > 256:
            value -= 256
        return value

    def _ensure_min_max(self, value: int) -> int:
        if self._min < 0 and value > 127:
            value -= 256

        if value < self._min or value > self._max:
            raise ValueError('value must between %d and %d, got %d' % (
                self._min,
                self._max,
                value
            ))
        else:
            return value


class OptionalInteger(CommandDataType[Optional[int]]):
    def __init__(self, _min: int, _max: int, _none: int) -> None:
        self._int = Integer(_min, _max)
        self._none = _none

    def to_typed(self, value: Any) -> Optional[int]:
        if value is None:
            return None
        elif value == self._none:
            return None
        else:
            return self._int.to_typed(value)

    def to_byte(self, value: Optional[int]) -> int:
        if value is None:
            return self._none
        else:
            return self._int.to_byte(value)


class IntegerEnum(CommandDataType[IntEnum]):
    def __init__(self, enum: Type[E]) -> None:
        self._enum = enum

    def to_typed(self, value: Any) -> E:
        if isinstance(value, self._enum):
            return value
        elif isinstance(value, int):
            return self._enum(value)
        elif isinstance(value, str):
            return self._enum[value.upper()]
        else:
            raise ValueError('cannot parse %s to %s' % (type(value), type(self._enum)))

    def to_byte(self, value: E) -> int:
        return value.value




from typing import Any, Optional
from unittest import TestCase

from tololib import ToloServer


class ServerTest(TestCase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._server: Optional[ToloServer] = None

    def setUp(self) -> None:
        self._server = ToloServer('localhost', 0)
        self._server.start()

    def tearDown(self) -> None:
        if self._server is not None:
            self._server.stop()
            self._server.join()
            self._server = None




from enum import IntEnum


DEFAULT_PORT = 51500

DEFAULT_RETRY_COUNT = 3
DEFAULT_RETRY_TIMEOUT = 1

MESSAGE_PREFIX = b'\xAA\xAA'
MESSAGE_SUFFIX = b'\x55\x55'


class AromaTherapy(IntEnum):
    A = 0
    B = 1


class Calefaction(IntEnum):
    HEAT = 0
    INACTIVE = 1
    UNCLEAR2 = 2  # TODO find correct meaning
    KEEP = 3


class LampMode(IntEnum):
    MANUAL = 0
    AUTOMATIC = 1


class Model(IntEnum):
    DOMESTIC = 0
    COMMERCIAL = 1




from .command import Command


SET_STATUS_COMMAND_PROPERTY_MAP = {
    Command.SET_POWER_ON: 'power_on',
    Command.SET_FAN_ON: 'fan_on',
    Command.SET_AROMA_THERAPY_ON: 'aroma_therapy_on',
    Command.SET_LAMP_ON: 'lamp_on',
    Command.SET_SWEEP_ON: 'sweep_on',
    Command.SET_SALT_BATH_ON: 'salt_bath_on'
}

GET_STATUS_COMMAND_PROPERTY_MAP = {
    Command.GET_FAN_TIMER: 'fan_timer',
    Command.GET_SALT_BATH_TIMER: 'salt_bath_timer'
}

SET_SETTINGS_COMMAND_PROPERTY_MAP = {
    Command.SET_TARGET_TEMPERATURE: 'target_temperature',
    Command.SET_TARGET_HUMIDITY: 'target_humidity',
    Command.SET_POWER_TIMER: 'power_timer',
    Command.SET_SALT_BATH_TIMER: 'salt_bath_timer',
    Command.SET_AROMA_THERAPY: 'aroma_therapy',
    Command.SET_SWEEP_TIMER: 'sweep_timer',
    Command.SET_LAMP_MODE: 'lamp_mode',
    Command.SET_FAN_TIMER: 'fan_timer'
}

GET_SETTINGS_COMMAND_PROPERTY_MAP = {
    Command.GET_LAMP_MODE: 'lamp_mode',
    Command.GET_AROMA_THERAPY: 'aroma_therapy'
}




import logging
import socket
from select import select
from typing import List, Optional, Tuple, Union

from .command import Command
from .const import AromaTherapy, LampMode, DEFAULT_RETRY_TIMEOUT, DEFAULT_RETRY_COUNT
from .message import Message
from .message_info import StatusInfo, SettingsInfo


logger = logging.getLogger(__name__)


class ToloClient(object):
    def __init__(self, address: str, port: int = 51500) -> None:
        self._address = address
        self._port = port

        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def send_message(self, message: Message) -> int:
        """Send a message to the server.

        Args:
            message (Message): The message to be send to the server.

        Returns:
            Number of bytes sent.
        """
        raw_bytes = bytes(message)
        logger.debug('sending %s to %s' % (raw_bytes.hex(), (self._address, self._port)))
        return self._socket.sendto(raw_bytes, (self._address, self._port))

    def receive_message(self, wait_timeout: Optional[float] = None) -> Optional[Message]:
        incoming = select([self._socket], [], [], wait_timeout)
        if incoming[0]:
            raw_bytes, sender = self._socket.recvfrom(4096)
            logger.debug('received %s from %s' % (raw_bytes.hex(), sender))
            return Message.from_bytes(raw_bytes)
        else:
            return None

    def send_wait_response(self, message: Message, resend_timeout: Optional[float] = None,
                           retries: int = 3) -> Optional[Message]:
        for _ in range(retries):
            self.send_message(message)
            response = self.receive_message(wait_timeout=resend_timeout)
            if response is not None:
                return response
        return None

    def get_status_info(self, resend_timeout: Optional[float] = None, retries: int = 3) -> Optional[StatusInfo]:
        response = self.send_wait_response(Message(Command.STATUS, 0, b'\xFF'), resend_timeout, retries)
        if response is None:
            return None
        return StatusInfo(response.info)

    def get_settings_info(self, resend_timeout: Optional[float] = None, retries: int = 3) -> Optional[SettingsInfo]:
        response = self.send_wait_response(Message(Command.SETTINGS, 0, b'\xFF'), resend_timeout, retries)
        if response is None:
            return None
        return SettingsInfo(response.info)

    @staticmethod
    def discover(address: str = '255.255.255.255', port: int = 51500,
                 wait_timeout: float = 1) -> List[Tuple[Message, Union[str, Tuple[str, int]]]]:
        devices = []

        discovery_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        discovery_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        discovery_socket.sendto(bytes(Message(Command.STATUS, 0, b'\xFF')), (address, port))

        while True:
            incoming_data = select([discovery_socket], [], [], wait_timeout)
            if incoming_data[0]:
                raw_bytes, sender = discovery_socket.recvfrom(4096)
                message = Message.from_bytes(raw_bytes)
                devices.append((message, sender))
            else:
                discovery_socket.close()
                return devices

    def close(self) -> None:
        self._socket.close()

    def set_power_on(self, power_on: bool, retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                     retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_POWER_ON,
            power_on,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_fan_on(self, fan_on: bool, retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                   retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_FAN_ON,
            fan_on,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_aroma_therapy_on(self, aroma_therapy_on: bool, retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                             retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_AROMA_THERAPY_ON,
            aroma_therapy_on,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_lamp_on(self, lamp_on: bool, retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                    retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_LAMP_ON,
            lamp_on,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_sweep_on(self, sweep_on: bool, retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                     retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_SWEEP_ON,
            sweep_on,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_salt_bath_on(self, salt_bath_on: bool, retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                         retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_SALT_BATH_ON,
            salt_bath_on,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_target_temperature(self, target_temperature: int, retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                               retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_TARGET_TEMPERATURE,
            target_temperature,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_target_humidity(self, target_humidity: int, retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                            retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_TARGET_HUMIDITY,
            target_humidity,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_power_timer(self, power_timer: Optional[int], retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                        retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_POWER_TIMER,
            power_timer,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_salt_bath_timer(self, salt_bath_timer: Optional[int],
                            retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                            retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_SALT_BATH_TIMER,
            salt_bath_timer,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_aroma_therapy(self, aroma_therapy: AromaTherapy, retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                          retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_AROMA_THERAPY,
            aroma_therapy,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_sweep_timer(self, sweep_timer: int, retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                        retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_SWEEP_TIMER,
            sweep_timer,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_lamp_mode(self, lamp_mode: LampMode, retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                      retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_LAMP_MODE,
            lamp_mode,
            b'\xFF'
        ), retry_timeout, retry_count)

    def set_fan_timer(self, fan_timer: Optional[int], retry_timeout: Optional[float] = DEFAULT_RETRY_TIMEOUT,
                      retry_count: int = DEFAULT_RETRY_COUNT) -> Optional[Message]:
        return self.send_wait_response(Message(
            Command.SET_FAN_TIMER,
            fan_timer,
            b'\xFF'
        ), retry_timeout, retry_count)

    def __del__(self) -> None:
        self.close()




from enum import Enum
from typing import Any, NamedTuple, Type, Union

from .const import AromaTherapy, LampMode
from .command_data_type import Boolean, Integer, OptionalInteger, IntegerEnum, CommandDataType

CommandTypeHint = Union[Type[int], Type[bool], int, None]


class CommandDefinition(NamedTuple):
    code: int
    default_value: Any
    command_data_type: CommandDataType[Any]


class Command(CommandDefinition, Enum):
    SET_TARGET_TEMPERATURE = CommandDefinition(4, 40, Integer(-20, 60))
    SET_POWER_TIMER = CommandDefinition(8, None, OptionalInteger(1, 60, 255))
    SET_POWER_ON = CommandDefinition(14, False, Boolean())
    SET_AROMA_THERAPY_ON = CommandDefinition(18, False, Boolean())
    SET_AROMA_THERAPY = CommandDefinition(20, AromaTherapy.A, IntegerEnum(AromaTherapy))
    GET_AROMA_THERAPY = CommandDefinition(21, AromaTherapy.A, IntegerEnum(AromaTherapy))
    GET_SWEEP_TIMER = CommandDefinition(26, False, Integer(0, 8))
    SET_SWEEP_TIMER = CommandDefinition(28, None, Integer(0, 8))
    SET_LAMP_ON = CommandDefinition(30, False, Boolean())
    SET_FAN_ON = CommandDefinition(34, False, Boolean())
    SET_FAN_TIMER = CommandDefinition(36, None, OptionalInteger(1, 60, 61))
    SET_TARGET_HUMIDITY = CommandDefinition(38, 95, Integer(60, 99))
    SET_SWEEP_ON = CommandDefinition(51, False, Boolean())
    GET_FAN_TIMER = CommandDefinition(53, None, OptionalInteger(0, 60, 61))
    SET_SALT_BATH_ON = CommandDefinition(54, False, Boolean())
    SET_SALT_BATH_TIMER = CommandDefinition(56, None, OptionalInteger(1, 60, 255))
    GET_SALT_BATH_TIMER = CommandDefinition(59, None, OptionalInteger(0, 100, 255))
    SET_LAMP_MODE = CommandDefinition(60, LampMode.MANUAL, IntegerEnum(LampMode))
    GET_LAMP_MODE = CommandDefinition(61, False, Boolean())
    LAMP_CHANGE_COLOR = CommandDefinition(62, 1, Integer(1, 1))
    STATUS = CommandDefinition(97, 0, Integer(0, 0))
    SETTINGS = CommandDefinition(99, 0, Integer(0, 0))

    @classmethod
    def from_code(cls, code: int) -> 'Command':
        for command in cls:
            if command.code == code:
                return command
        else:
            raise ValueError('unknown command code %s' % code)




import argparse
import logging
import signal
import sys
from types import FrameType

from .client import ToloClient
from .command import Command
from .const import DEFAULT_PORT
from .message_info import SettingsInfo, StatusInfo
from .message import Message
from .server import ToloServer
from .simulator import Simulator


logger = logging.getLogger(__name__)

SETTINGS_COMMAND_MAP = {
    # main panel
    'power': Command.SET_POWER_ON,
    'fan': Command.SET_FAN_ON,
    'aroma-therapy': Command.SET_AROMA_THERAPY_ON,
    'lamp': Command.SET_LAMP_ON,
    'sweep': Command.GET_SWEEP_TIMER,
    'salt-bath': Command.SET_SALT_BATH_TIMER,
    # settings panel
    'target-temperature': Command.SET_TARGET_TEMPERATURE,
    'target-humidity': Command.SET_TARGET_HUMIDITY,
    'power-timer': Command.SET_POWER_TIMER,
    'salt-bath-timer': Command.SET_SALT_BATH_TIMER,
    'aroma-therapy-slot': Command.SET_AROMA_THERAPY,
    'sweep-timer': Command.SET_SWEEP_TIMER,
    'lamp-mode': Command.SET_LAMP_MODE,
    'fan-timer': Command.SET_FAN_TIMER
}


def main() -> int:
    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument('-l', '--log-level', choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'),
                                 default='INFO')

    argument_sub_parser = argument_parser.add_subparsers(title='command', dest='command', required=True)

    server_argument_parser = argument_sub_parser.add_parser('server')
    server_argument_parser.add_argument('-l', '--listen-address', default='')
    server_argument_parser.add_argument('-p', '--port', type=int, default=DEFAULT_PORT)
    server_argument_parser.add_argument('-s', '--simulate', action='store_true')

    discover_argument_parser = argument_sub_parser.add_parser('discover')
    discover_argument_parser.add_argument('-p', '--port', type=int, default=DEFAULT_PORT)

    show_settings_argument_parser = argument_sub_parser.add_parser('show-settings')
    show_settings_argument_parser.add_argument('-a', '--address', type=str)
    show_settings_argument_parser.add_argument('-p', '--port', type=int, default=DEFAULT_PORT)
    show_settings_argument_parser.add_argument('--resend-timeout', type=float, default=1)
    show_settings_argument_parser.add_argument('--retries', type=int, default=3)

    set_argument_parser = argument_sub_parser.add_parser('set')
    set_argument_parser.add_argument('-a', '--address', type=str)
    set_argument_parser.add_argument('-p', '--port', type=int, default=DEFAULT_PORT)
    set_argument_parser.add_argument('--resend-timeout', type=float, default=1)
    set_argument_parser.add_argument('--retries', type=int, default=3)
    set_argument_parser.add_argument('key', choices=SETTINGS_COMMAND_MAP.keys())
    set_argument_parser.add_argument('value')

    args = argument_parser.parse_args()

    library_logger = logging.getLogger(__package__)
    log_handler = logging.StreamHandler(sys.stderr)
    log_formatter = logging.Formatter(fmt='%(asctime)s %(name)-20s [%(levelname)s] %(message)s')
    log_handler.setFormatter(log_formatter)
    library_logger.addHandler(log_handler)
    library_logger.setLevel(logging.getLevelName(args.log_level))

    if args.command == 'server':
        return cmd_server(args)
    elif args.command == 'discover':
        return cmd_discover(args)
    elif args.command == 'show-settings':
        return cmd_show_settings(args)
    elif args.command == 'set':
        return cmd_set(args)
    else:
        return 1


def cmd_server(args: argparse.Namespace) -> int:
    server = ToloServer(address=args.listen_address, port=args.port)

    def stop_on_signal(s: signal.Signals, t: FrameType) -> None:
        logger.debug('received signal %d, shutting down...' % s)
        server.stop()

    signal.signal(signal.SIGINT, stop_on_signal)

    if args.simulate:
        simulator = Simulator(server)
        simulator.start()

    server.run()
    return 0


def cmd_discover(args: argparse.Namespace) -> int:
    # TODO use args
    for message, address in ToloClient.discover():
        info = StatusInfo(message.info)
        print('=== Address: %s:%s' % (address[0], address[1]))
        print(info)
    return 0


def cmd_show_settings(args: argparse.Namespace) -> int:
    message = Message(Command.SETTINGS, Command.SETTINGS.default_value, b'\xFF')
    client = ToloClient(args.address, args.port)
    response = client.send_wait_response(message, resend_timeout=args.resend_timeout, retries=args.retries)
    if response is None:
        print('Did not get response')
        return 1
    else:
        info = SettingsInfo(response.info)
        print(info)
        return 0


def cmd_set(args: argparse.Namespace) -> int:
    command = SETTINGS_COMMAND_MAP.get(args.key)
    if command is None:
        raise ValueError('command not supported')  # TODO change error type

    data = command.command_data_type.to_byte_any(args.value)

    message = Message(command, data, b'\xFF')

    client = ToloClient(args.address, args.port)
    response = client.send_wait_response(message, resend_timeout=args.resend_timeout, retries=args.retries)
    if response is None:
        return 1
    else:
        print(response)
        return 0




from typing import Any, Optional
from unittest import TestCase

from tololib import ToloClient, ToloServer
from tololib.command import Command
from tololib.const import AromaTherapy, LampMode
from tololib.message_info import StatusInfo, SettingsInfo


class ClientClassTest(TestCase):
    def test_discovery(self) -> None:
        self._server = ToloServer('', 0)
        self._server.start()
        result = ToloClient.discover(port=self._server.get_socket_info()[1])
        self.assertEqual(len(result), 1)
        self._server.stop()


class ClientInstanceTest(TestCase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._server: Optional[ToloServer] = None

    def setUp(self) -> None:
        self._server = ToloServer('localhost', 0)
        self._server.start()

    def tearDown(self) -> None:
        if self._server is not None:
            self._server.stop()
            self._server.join()
            self._server = None

    def _get_port(self) -> int:
        if self._server is None:
            raise RuntimeError('server not set up')

        port = self._server.get_socket_info()[1]

        if not isinstance(port, int):
            raise RuntimeError('not an int value')

        return port

    def test_get_status_info(self) -> None:
        client = ToloClient('localhost', self._get_port())
        response = client.get_status_info()
        self.assertIsInstance(response, StatusInfo)

    def test_get_settings_info(self) -> None:
        client = ToloClient('localhost', self._get_port())
        response = client.get_settings_info()
        self.assertIsInstance(response, SettingsInfo)

    def test_main_setters(self) -> None:
        client = ToloClient('localhost', self._get_port())

        response = client.set_power_on(True)
        assert response is not None
        self.assertEqual(response.command, Command.SET_POWER_ON)
        self.assertEqual(response.data, True)

        response = client.set_power_on(False)
        assert response is not None
        self.assertEqual(response.command, Command.SET_POWER_ON)
        self.assertEqual(response.data, False)

        response = client.set_fan_on(True)
        assert response is not None
        self.assertEqual(response.command, Command.SET_FAN_ON)
        self.assertEqual(response.data, True)

        response = client.set_fan_on(False)
        assert response is not None
        self.assertEqual(response.command, Command.SET_FAN_ON)
        self.assertEqual(response.data, False)

        response = client.set_aroma_therapy_on(True)
        assert response is not None
        self.assertEqual(response.command, Command.SET_AROMA_THERAPY_ON)
        self.assertEqual(response.data, True)

        response = client.set_aroma_therapy_on(False)
        assert response is not None
        self.assertEqual(response.command, Command.SET_AROMA_THERAPY_ON)
        self.assertEqual(response.data, False)

        response = client.set_lamp_on(True)
        assert response is not None
        self.assertEqual(response.command, Command.SET_LAMP_ON)
        self.assertEqual(response.data, True)

        response = client.set_lamp_on(False)
        assert response is not None
        self.assertEqual(response.command, Command.SET_LAMP_ON)
        self.assertEqual(response.data, False)

        response = client.set_sweep_on(True)
        assert response is not None
        self.assertEqual(response.command, Command.SET_SWEEP_ON)
        self.assertEqual(response.data, True)

        response = client.set_sweep_on(False)
        assert response is not None
        self.assertEqual(response.command, Command.SET_SWEEP_ON)
        self.assertEqual(response.data, False)

        response = client.set_salt_bath_on(True)
        assert response is not None
        self.assertEqual(response.command, Command.SET_SALT_BATH_ON)
        self.assertEqual(response.data, True)

        response = client.set_salt_bath_on(False)
        assert response is not None
        self.assertEqual(response.command, Command.SET_SALT_BATH_ON)
        self.assertEqual(response.data, False)

    def test_setting_setters(self) -> None:
        client = ToloClient('localhost', self._get_port())

        # target temperature
        response = client.set_target_temperature(45)
        assert response is not None
        self.assertEqual(response.command, Command.SET_TARGET_TEMPERATURE)
        self.assertEqual(response.data, 45)

        self.assertRaises(ValueError, client.set_target_temperature, 90)
        self.assertRaises(ValueError, client.set_target_temperature, -33)
        self.assertRaises(ValueError, client.set_target_temperature, None)

        # target humidity
        response = client.set_target_humidity(93)
        assert response is not None
        self.assertEqual(response.command, Command.SET_TARGET_HUMIDITY)
        self.assertEqual(response.data, 93)

        self.assertRaises(ValueError, client.set_target_humidity, -3)
        self.assertRaises(ValueError, client.set_target_humidity, 101)
        self.assertRaises(ValueError, client.set_target_humidity, None)

        # power timer
        response = client.set_power_timer(None)
        assert response is not None
        self.assertEqual(response.command, Command.SET_POWER_TIMER)
        self.assertEqual(response.data, None)

        response = client.set_power_timer(30)
        assert response is not None
        self.assertEqual(response.command, Command.SET_POWER_TIMER)
        self.assertEqual(response.data, 30)

        self.assertRaises(ValueError, client.set_power_timer, -1)
        self.assertRaises(ValueError, client.set_power_timer, 0)
        self.assertRaises(ValueError, client.set_power_timer, 61)

        # salt bath timer
        response = client.set_salt_bath_timer(None)
        assert response is not None
        self.assertEqual(response.command, Command.SET_SALT_BATH_TIMER)
        self.assertEqual(response.data, None)

        response = client.set_salt_bath_timer(30)
        assert response is not None
        self.assertEqual(response.command, Command.SET_SALT_BATH_TIMER)
        self.assertEqual(response.data, 30)

        self.assertRaises(ValueError, client.set_salt_bath_timer, 0)
        self.assertRaises(ValueError, client.set_salt_bath_timer, -1)
        self.assertRaises(ValueError, client.set_salt_bath_timer, 61)

        # aroma therapy
        for aroma_therapy in AromaTherapy:
            response = client.set_aroma_therapy(aroma_therapy)
            assert response is not None
            self.assertEqual(response.command, Command.SET_AROMA_THERAPY)
            self.assertEqual(response.data, aroma_therapy)

        self.assertRaises(ValueError, client.set_aroma_therapy, -1)
        self.assertRaises(ValueError, client.set_aroma_therapy, 2)
        self.assertRaises(ValueError, client.set_aroma_therapy, None)

        # sweep timer
        response = client.set_sweep_timer(5)
        assert response is not None
        self.assertEqual(response.command, Command.SET_SWEEP_TIMER)
        self.assertEqual(response.data, 5)

        self.assertRaises(ValueError, client.set_sweep_timer, -1)
        self.assertRaises(ValueError, client.set_sweep_timer, 9)
        self.assertRaises(ValueError, client.set_sweep_timer, None)

        # lamp mode
        for lamp_mode in LampMode:
            response = client.set_lamp_mode(lamp_mode)
            assert response is not None
            self.assertEqual(response.command, Command.SET_LAMP_MODE)
            self.assertEqual(response.data, lamp_mode)

        self.assertRaises(ValueError, client.set_lamp_mode, None)
        self.assertRaises(ValueError, client.set_lamp_mode, -1)
        self.assertRaises(ValueError, client.set_lamp_mode, 2)

        # fan timer
        response = client.set_fan_timer(30)
        assert response is not None
        self.assertEqual(response.command, Command.SET_FAN_TIMER)
        self.assertEqual(response.data, 30)

        response = client.set_fan_timer(None)
        assert response is not None
        self.assertEqual(response.command, Command.SET_FAN_TIMER)
        self.assertEqual(response.data, None)

        response = client.set_fan_timer(61)
        assert response is not None
        self.assertEqual(response.command, Command.SET_FAN_TIMER)
        self.assertEqual(response.data, None)

        self.assertRaises(ValueError, client.set_fan_timer, 0)
        self.assertRaises(ValueError, client.set_fan_timer, -1)




import logging
from threading import Thread
from time import sleep

from .const import AromaTherapy, Calefaction, LampMode
from .server import ToloServer


logger = logging.getLogger(__name__)


class Simulator(Thread):
    def __init__(self, server: ToloServer) -> None:
        super().__init__(daemon=True)
        self._server = server

    def run(self) -> None:
        logger.debug('simulator started')

        # assign shortcut variables
        status = self._server.status
        settings = self._server.settings

        # define reasonable default values
        status.calefaction = Calefaction.INACTIVE
        status.flow_in = False
        status.flow_out = True
        status.current_humidity = 25
        status.tank_temperature = 15

        settings.target_temperature = 45
        settings.target_humidity = 95
        settings.power_timer = 30
        settings.salt_bath_timer = None
        settings.aroma_therapy = AromaTherapy.B
        settings.sweep_timer = 0
        settings.lamp_mode = LampMode.MANUAL
        settings.fan_timer = 10

        # the loop
        while True:
            if status.power_on:
                status.flow_out = False
                if status.water_level < 3:
                    status.flow_in = True
                    status.water_level += 1
                else:
                    status.flow_in = False

                if status.tank_temperature < 100:
                    status.calefaction = Calefaction.HEAT
                    status.tank_temperature += 5
                else:
                    status.calefaction = Calefaction.KEEP

            else:
                status.flow_in = False
                status.calefaction = Calefaction.INACTIVE
                if status.water_level > 0:
                    status.flow_out = True
                    status.water_level -= 1

            sleep(1)




from typing import Optional, cast

from .command import Command
from .command_data_type import Integer, IntegerEnum
from .const import AromaTherapy, Calefaction, LampMode, Model


class MessageInfo(bytearray):
    LENGTH = -1

    def __init__(self, initial_value: Optional[bytes] = None) -> None:
        if initial_value is None:
            super().__init__(b'\x00' * self.LENGTH)
        elif len(initial_value) == self.LENGTH:
            super().__init__(initial_value)
        else:
            raise ValueError('initial value must have length %d, %d given' % (self.LENGTH, len(initial_value)))


class StatusInfo(MessageInfo):
    """
      * 0: int(self._power_on),
      * 1: self._current_temperature,
      * 2: 61 if self._power_timer is None else self._power_timer,
      * 3: (64 if self._flow_in else 0) + (16 if self._flow_out else 0) + self._calefaction.value,
     4: int(self._aroma_therapy_on),
     5: int(self._sweep_on),
     6: 0,  # TODO descaling - what is this?
     7: int(self._lamp_on),
     8: self._water_level,
     9: int(self._fan_on),
    10: 61 if self._fan_timer is None else self._fan_timer,
    11: self._current_humidity,
    12: self._tank_temperature,
    13: 0,  # TODO unused?
    14: self._model.value,
    15: self._salt_bath_on,
    16: 0 if self._salt_bath_timer is None else self._salt_bath_timer
    """
    LENGTH = 17

    @property
    def power_on(self) -> bool:
        return cast(bool, Command.SET_POWER_ON.command_data_type.to_typed(self[0]))

    @power_on.setter
    def power_on(self, power_on: bool) -> None:
        self[0] = Command.SET_POWER_ON.command_data_type.to_byte(power_on)

    @property
    def current_temperature(self) -> int:
        return cast(int, Command.SET_TARGET_TEMPERATURE.command_data_type.to_typed(self[1]))

    @current_temperature.setter
    def current_temperature(self, current_temperature: int) -> None:
        self[1] = Command.SET_TARGET_TEMPERATURE.command_data_type.to_byte(current_temperature)

    @property
    def power_timer(self) -> Optional[int]:
        return cast(Optional[int], Command.SET_POWER_TIMER.command_data_type.to_typed(self[2]))

    @power_timer.setter
    def power_timer(self, power_timer: Optional[int]) -> None:
        self[2] = Command.SET_POWER_TIMER.command_data_type.to_byte(power_timer)

    @property
    def flow_in(self) -> bool:
        return bool(self[3] & 64)

    @flow_in.setter
    def flow_in(self, flow_in: bool) -> None:
        self[3] = (self[3] & 191) | (64 if flow_in else 0)

    @property
    def flow_out(self) -> bool:
        return bool(self[3] & 16)

    @flow_out.setter
    def flow_out(self, flow_out: bool) -> None:
        self[3] = (self[3] & 239) | (16 if flow_out else 0)

    @property
    def calefaction(self) -> Calefaction:
        return Calefaction(self[3] & 3)

    @calefaction.setter
    def calefaction(self, calefaction: Calefaction) -> None:
        self[3] = (self[3] & 252) | IntegerEnum(Calefaction).to_byte(calefaction)

    @property
    def aroma_therapy_on(self) -> bool:
        return cast(bool, Command.SET_AROMA_THERAPY_ON.command_data_type.to_typed(self[4]))

    @aroma_therapy_on.setter
    def aroma_therapy_on(self, aroma_therapy_on: bool) -> None:
        self[4] = Command.SET_AROMA_THERAPY_ON.command_data_type.to_byte(aroma_therapy_on)

    @property
    def sweep_on(self) -> bool:
        return cast(bool, Command.GET_SWEEP_TIMER.command_data_type.to_typed(self[5]))

    @sweep_on.setter
    def sweep_on(self, sweep_on: bool) -> None:
        self[5] = Command.GET_SWEEP_TIMER.command_data_type.to_byte(sweep_on)

    @property
    def lamp_on(self) -> bool:
        return cast(bool, Command.SET_LAMP_ON.command_data_type.to_typed(self[7]))

    @lamp_on.setter
    def lamp_on(self, lamp_on: bool) -> None:
        self[7] = Command.SET_LAMP_ON.command_data_type.to_byte(lamp_on)

    @property
    def water_level(self) -> int:
        return Integer(0, 3).to_typed(self[8])

    @water_level.setter
    def water_level(self, water_level: int) -> None:
        self[8] = Integer(0, 3).to_byte(water_level)

    @property
    def fan_on(self) -> bool:
        return cast(bool, Command.SET_FAN_ON.command_data_type.to_typed(self[9]))

    @fan_on.setter
    def fan_on(self, fan_on: bool) -> None:
        self[9] = Command.SET_FAN_ON.command_data_type.to_byte(fan_on)

    @property
    def fan_timer(self) -> Optional[int]:
        return cast(Optional[int], Command.GET_FAN_TIMER.command_data_type.to_typed(self[10]))

    @fan_timer.setter
    def fan_timer(self, fan_timer: Optional[int]) -> None:
        self[10] = Command.SET_FAN_TIMER.command_data_type.to_byte(fan_timer)

    @property
    def current_humidity(self) -> int:
        return cast(int, Command.SET_TARGET_HUMIDITY.command_data_type.to_typed(self[11]))

    @current_humidity.setter
    def current_humidity(self, current_humidity: int) -> None:
        self[11] = Command.SET_TARGET_HUMIDITY.command_data_type.to_byte(current_humidity)

    @property
    def tank_temperature(self) -> int:
        return Integer(0, 255).to_typed(self[12])

    @tank_temperature.setter
    def tank_temperature(self, tank_temperature: int) -> None:
        self[12] = Integer(0, 255).to_byte(tank_temperature)

    @property
    def model(self) -> Model:
        return IntegerEnum(Model).to_typed(self[14])

    @model.setter
    def model(self, model: Model) -> None:
        self[14] = IntegerEnum(Model).to_byte(model)

    @property
    def salt_bath_on(self) -> bool:
        return cast(bool, Command.SET_SALT_BATH_ON.command_data_type.to_typed(self[15]))

    @salt_bath_on.setter
    def salt_bath_on(self, salt_bath_on: bool) -> None:
        self[15] = Command.SET_SALT_BATH_ON.command_data_type.to_byte(salt_bath_on)

    @property
    def salt_bath_timer(self) -> Optional[int]:
        return cast(Optional[int], Command.SET_SALT_BATH_TIMER.command_data_type.to_typed(self[16]))

    @salt_bath_timer.setter
    def salt_bath_timer(self, salt_bath_timer: Optional[int]) -> None:
        self[16] = Command.SET_SALT_BATH_TIMER.command_data_type.to_byte(salt_bath_timer)

    def __str__(self) -> str:
        return '\n'.join([
            'Power:               %s' % ('ON' if self.power_on else 'OFF'),
            'Current Temperature: %d' % self.current_temperature,
            'Power Timer:         %s' % self.power_timer,
            'Flow in:             %s' % ('YES' if self.flow_in else 'NO'),
            'Flow out:            %s' % ('YES' if self.flow_out else 'NO'),
            'Calefaction:         %s' % self.calefaction.name,
            'Aroma Therapy:       %s' % ('ON' if self.aroma_therapy_on else 'OFF'),
            'Sweep:               %s' % ('ON' if self.sweep_on else 'OFF'),
            # 'Descaling:           %s' % ('YES' if self.descaling else 'NO'), TODO what is this?
            'Lamp:                %s' % ('ON' if self.lamp_on else 'OFF'),
            'Water Level:         %s' % self.water_level,
            'Fan:                 %s' % ('ON' if self.fan_on else 'OFF'),
            'Fan Timer:           %s' % ('OFF' if self.fan_timer is None else str(self.fan_timer)),
            'Current Humidity:    %s' % self.current_humidity,
            'Tank Temperature:    %s' % self.tank_temperature,
            'Model:               %s' % self.model.name,
            'Salt Bath On:        %s' % ('ON' if self.salt_bath_on else 'OFF'),
            'Salt Bath Timer:     %s' % ('OFF' if self.salt_bath_timer is None else str(self.salt_bath_timer))
        ])


class SettingsInfo(MessageInfo):
    """
    0: self._target_temperature,
    1: 255 if self._power_timer is None else self._power_timer,
    2: self._aroma_therapy.value,
    3: self._sweep_timer,
    4: 61 if self._fan_timer is None else self._fan_timer,
    5: self._target_humidity,
    6: 255 if self._salt_bath_timer is None else self._salt_bath_timer,
    7: self._lamp_mode.value

    """
    LENGTH = 8

    @property
    def target_temperature(self) -> int:
        return cast(int, Command.SET_TARGET_TEMPERATURE.command_data_type.to_typed(self[0]))

    @target_temperature.setter
    def target_temperature(self, target_temperature: int) -> None:
        self[0] = Command.SET_TARGET_TEMPERATURE.command_data_type.to_byte(target_temperature)

    @property
    def power_timer(self) -> Optional[int]:
        return cast(Optional[int], Command.SET_POWER_TIMER.command_data_type.to_typed(self[1]))

    @power_timer.setter
    def power_timer(self, power_timer: Optional[int]) -> None:
        self[1] = Command.SET_POWER_TIMER.command_data_type.to_byte(power_timer)

    @property
    def aroma_therapy(self) -> AromaTherapy:
        return cast(AromaTherapy, Command.SET_AROMA_THERAPY.command_data_type.to_typed(self[2]))

    @aroma_therapy.setter
    def aroma_therapy(self, aroma_therapy: AromaTherapy) -> None:
        self[2] = Command.SET_AROMA_THERAPY.command_data_type.to_byte(aroma_therapy)

    @property
    def sweep_timer(self) -> int:
        return cast(int, Command.SET_SWEEP_TIMER.command_data_type.to_typed(self[3]))

    @sweep_timer.setter
    def sweep_timer(self, sweep_timer: int) -> None:
        self[3] = Command.SET_SWEEP_TIMER.command_data_type.to_byte(sweep_timer)

    @property
    def fan_timer(self) -> Optional[int]:
        return cast(Optional[int], Command.SET_FAN_TIMER.command_data_type.to_typed(self[4]))

    @fan_timer.setter
    def fan_timer(self, fan_timer: Optional[int]) -> None:
        self[4] = Command.SET_FAN_TIMER.command_data_type.to_byte(fan_timer)

    @property
    def target_humidity(self) -> int:
        return cast(int, Command.SET_TARGET_HUMIDITY.command_data_type.to_typed(self[5]))

    @target_humidity.setter
    def target_humidity(self, target_humidity: int) -> None:
        self[5] = Command.SET_TARGET_HUMIDITY.command_data_type.to_byte(target_humidity)

    @property
    def salt_bath_timer(self) -> Optional[int]:
        return cast(Optional[int], Command.SET_SALT_BATH_TIMER.command_data_type.to_typed(self[6]))

    @salt_bath_timer.setter
    def salt_bath_timer(self, salt_bath_timer: Optional[int]) -> None:
        self[6] = Command.SET_SALT_BATH_TIMER.command_data_type.to_byte(salt_bath_timer)

    @property
    def lamp_mode(self) -> LampMode:
        return cast(LampMode, Command.SET_LAMP_MODE.command_data_type.to_typed(self[7]))

    @lamp_mode.setter
    def lamp_mode(self, lamp_mode: LampMode) -> None:
        self[7] = Command.SET_LAMP_MODE.command_data_type.to_byte(lamp_mode)

    def __str__(self) -> str:
        return '\n'.join([
            'Target Temperature: %d' % self.target_temperature,
            'Power Timer:        %s' % self.power_timer,
            'Aroma Therapy:      %s' % self.aroma_therapy,
            'Sweep Timer:        %s' % self.sweep_timer,
            'Fan Timer:          %s' % self.fan_timer,
            'Target Humidity:    %s' % self.target_humidity,
            'Salt Bath Timer:    %s' % self.salt_bath_timer,
            'Lamp Mode:          %s' % self.lamp_mode
        ])




from unittest import TestCase

from tololib import Message


class MessageTest(TestCase):
    def test_init(self) -> None:
        pass  # TODO write tests

    def test_generate_crc(self) -> None:
        self.assertEqual(Message.generate_crc(b'\xaa\xaaa\x00\xffUU'), 0x9e)

        self.assertNotEqual(Message.generate_crc(b'\xaa\xaaa\x00\xffUU'), 0)

    def test_validate_crc(self) -> None:
        self.assertTrue(Message.validate_crc(b'\xaa\xaaa\x00\xffUU\x9e'))

        self.assertFalse(Message.validate_crc(b'\xaa\xaaa\x00\xffUU\x00'))

    def test_validate_meta(self) -> None:
        self.assertTrue(Message.validate_meta(b'\xaa\xaaa\x00\xffUU\x9e'))
        self.assertTrue(Message.validate_meta(b'\xaa\xaaFOOBAR\xffUU\xe8'))

        self.assertFalse(Message.validate_meta(b'\xaa\xaaa\x00\xffUU\x00'))
        self.assertFalse(Message.validate_meta(b'\xaaa\x00\xffU\x9e'))

    def test_to_bytes(self) -> None:
        pass  # TODO write tests




from setuptools import setup, find_packages


with open('README.md', 'r') as fp:
    long_description = fp.read()

setup(
    name='tololib',
    description='Python Library for Controlling TOLO Sauna/Steam Bath Devices',
    long_description=long_description,
    long_description_content_type='text/markdown',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author='Matthias Lohr',
    author_email='mail@mlohr.com',
    url='https://gitlab.com/MatthiasLohr/tololib',
    license='MIT',
    install_requires=[],
    python_requires='>=3.7.*, <4',
    packages=find_packages(exclude=['tests', 'tests.*']),
    package_data={
        'tololib': ['py.typed']
    },
    entry_points={
        'console_scripts': [
            'tolo-cli=tololib.cli:main'
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Home Automation',
        'Typing :: Typed'
    ],
    project_urls={
        'Documentation': 'https://matthiaslohr.gitlab.io/tololib/',
        'Source': 'https://gitlab.com/MatthiasLohr/tololib',
        'Tracker': 'https://gitlab.com/MatthiasLohr/tololib/issues'
    }
)
